/* KEYBOARD.C
 * Provides low level keyboard access for programming games.
 *
 * by Gary Neal, Jr. -- garyneal@mailcity.com
 * http://www.geocities.com/SiliconValley/Park/7113/
 *
 * Last revision: February 19, 2000
 * Parts of this work was based on other works by:
 *     Dave Roberts:  PC Game Programming Explorer
 *     Andre LaMothe:  Teach Yourself Game Programming in 21 Days
 *     Shawn Hargreaves and the Allegro Team
 */

#include "keyboard.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Table of boolean values for checking and recording the keyboard state */
static volatile unsigned char keyIsPressed[128];
static volatile unsigned char keyWasPressed[128];

/* Flags a CTRL-ALT-DEL or CTRL-ALT-ESC key sequence */
static volatile bool keyHardAbort = false;

/* Flags whether low level keyboard handler should chain BIOS's handler */
static bool chainBiosIsr = false;

/* Flags whether low level keyboard handler should call a second function */
static bool callNextFunc = false;

/* ISR Table used by the handler */
#define isrTable ((long far *)(MK_FP(0, 0)))

/* Stores the address of the BIOS's interrupt handler */
#ifdef __cplusplus
static volatile void interrupt (far *OldKeyboardHandler)(...) = 0;
#else
static volatile void interrupt (far *OldKeyboardHandler)() = 0;
#endif

/* Stores the address of the chained function */
static void (far *nextFunc)(void) = 0;

/* The interrupt handler used by the keyboard functions */
#ifdef __cplusplus
static void interrupt far KeyboardInterruptHandler(...)
#else
static void interrupt far KeyboardInterruptHandler()
#endif
{
    static bool extended = false;
    register int i;

    enable();                    /* Immediately enable interrupts */
    i = inp(0x60);               /* Read the scan code */

    /* Check the scan code and update the keyboard tables */
    if (i == 0xE0)
        extended = true;
    else {
        if (i & 0x80) {
            i &= 0x7F;
            if (extended)
                keyIsPressed[i] &= 0x7F;
            else
                keyIsPressed[i] &= 0xFE;
        } else {
            if (extended) {
                keyIsPressed[i]  |= 0x80;
                keyWasPressed[i] |= 0x80;
            } else {
                keyIsPressed[i]  |= 0x01;
                keyWasPressed[i] |= 0x01;
            }
        }
        extended = false;
    }

    /* Check to see if a second function is being chained */
    if (callNextFunc)
        nextFunc();

    /* Check for CTRL-ALT-DEL or CTRL-ALT-ESC sequences */
    if (keyIsPressed[KEY_CTRL] && keyIsPressed[KEY_ALT] &&
        (keyIsPressed[KEY_DEL] || keyIsPressed[KEY_ESC]))
    {
        disable();                /* Disable interrupts */

        /* Restore BIOS's keyboard handler */
        isrTable[9] = (long)OldKeyboardHandler;

        enable();                 /* Enable interrupts */
        OldKeyboardHandler = 0;   /* Remove the low level keyboard handler */
        keyHardAbort = true;      /* Flag a hard abort */
    } else if (chainBiosIsr)
        _chain_intr(OldKeyboardHandler);

    /* Clear the keyboard controller */
    i = inp(0x61);
    outp(0x61, i | 0x80);
    outp(0x61, i);

    outp(0x20, 0x20);             /* Flag an end of interrupt to the PIC */
}

/* Installs the low level keyboard handler for playing games */
void InstallKeyboardHandler(bool chainBios, void (far *func)(void))
{
    int i;

    /* Set up flags and clear the keyboard state table */
    keyHardAbort = false;
    chainBiosIsr = chainBios;
    callNextFunc = (bool)((nextFunc = func) != 0);
    for (i = 0; i < 128; ++i)
        keyIsPressed[i] = keyWasPressed[i] = 0;

    if (OldKeyboardHandler != 0) return;

    /* Get BIOS's keyboard handler and store it */
    OldKeyboardHandler = getvect(9);

    /* Set the interrupt vector to the low level keyboard handler */
    setvect(9, KeyboardInterruptHandler);
}

/* Removes the low level keyboard handler */
void RemoveKeyboardHandler(void)
{
    /* Return if low level keyboard handler is not installed */
    if (OldKeyboardHandler == 0) return;

    /* Restores BIOS's keyboard handler */
    setvect(9, OldKeyboardHandler);
    OldKeyboardHandler = 0;       /* Remove the low level keyboard handler */
}

/* Checks to see if a key was pressed since a previous call to
 * KeyPressed or KeyPressedNow
 */
int KeyPressed(int c)
{
    bool retVal;

    c &= 0x7F;                 /* Ensure correct scan code value */

    /* Get the key state */
    retVal = keyIsPressed[c] | keyWasPressed[c];
    keyWasPressed[c] = 0;      /* Clear the 'was pressed' state */
    return retVal;             /* Return the key state */
}

/* Checks to see if a key is currently being pressed. */
int KeyPressedNow(int c)
{
    c &= 0x7F;                 /* Ensure correct scan code value */
    keyWasPressed[c] = 0;      /* Clear the 'was pressed' state */
    return keyIsPressed[c];    /* Return the 'is pressed' state */
}

/* Returns the hard abort flag */
bool KeyHardAbort(void)
{
    return keyHardAbort;       /* Return the hard abort value */
}

#ifdef __cplusplus
}
#endif
